@@ -30,6 +30,8 @@
  *	surface transform disable key: r
  */
 
+#include "config.h"
+
 #include <stdint.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -791,18 +793,14 @@ cliptest_create(struct display *display)
 {
 	struct cliptest *cliptest;
 
-	cliptest = malloc(sizeof *cliptest);
-	if (cliptest == NULL)
-		return cliptest;
-	memset(cliptest, 0, sizeof *cliptest);
-
+	cliptest = xzalloc(sizeof *cliptest);
 	cliptest->surface.geometry = &cliptest->geometry;
 	cliptest->surface.transform.enabled = 0;
 	geometry_init(&cliptest->geometry);
 	geometry_init(&cliptest->ui.geometry);
 
 	cliptest->window = window_create(display);
-	cliptest->widget = frame_create(cliptest->window, cliptest);
+	cliptest->widget = window_frame_create(cliptest->window, cliptest);
 	window_set_title(cliptest->window, "cliptest");
 	cliptest->display = display;
 
@@ -890,7 +888,7 @@ main(int argc, char *argv[])
 	if (argc > 1)
 		return benchmark();
 
-	d = display_create(argc, argv);
+	d = display_create(&argc, argv);
 	if (d == NULL) {
 		fprintf(stderr, "failed to create display: %m\n");
 		return -1;
