@@ -401,11 +401,10 @@ gears_create(struct display *display)
 	struct timeval tv;
 	int i;
 
-	gears = malloc(sizeof *gears);
-	memset(gears, 0, sizeof *gears);
+	gears = zalloc(sizeof *gears);
 	gears->d = display;
 	gears->window = window_create(display);
-	gears->widget = frame_create(gears->window, gears);
+	gears->widget = window_frame_create(gears->window, gears);
 	window_set_title(gears->window, "Wayland Gears");
 
 	gears->display = display_get_egl_display(gears->d);
@@ -474,7 +473,7 @@ int main(int argc, char *argv[])
 {
 	struct display *d;
 
-	d = display_create(argc, argv);
+	d = display_create(&argc, argv);
 	if (d == NULL) {
 		fprintf(stderr, "failed to create display: %m\n");
 		return -1;
