@@ -21,6 +21,8 @@
  * OF THIS SOFTWARE.
  */
 
+#include "config.h"
+
 #include <stdint.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -336,7 +338,7 @@ fullscreen_handler(struct window *window, void *data)
 }
 
 static void
-close_handler(struct window *window, void *data)
+close_handler(void *data)
 {
 	struct image *image = data;
 
@@ -358,10 +360,9 @@ image_create(struct display *display, const char *filename,
 	struct image *image;
 	char *b, *copy, title[512];;
 
-	image = malloc(sizeof *image);
+	image = zalloc(sizeof *image);
 	if (image == NULL)
 		return image;
-	memset(image, 0, sizeof *image);
 
 	copy = strdup(filename);
 	b = basename(copy);
@@ -373,11 +374,13 @@ image_create(struct display *display, const char *filename,
 
 	if (!image->image) {
 		fprintf(stderr, "could not find the image %s!\n", b);
+		free(image->filename);
+		free(image);
 		return NULL;
 	}
 
 	image->window = window_create(display);
-	image->widget = frame_create(image->window, image);
+	image->widget = window_frame_create(image->window, image);
 	window_set_title(image->window, title);
 	image->display = display;
 	image->image_counter = image_counter;
@@ -409,7 +412,7 @@ main(int argc, char *argv[])
 	int i;
 	int image_counter = 0;
 
-	d = display_create(argc, argv);
+	d = display_create(&argc, argv);
 	if (d == NULL) {
 		fprintf(stderr, "failed to create display: %m\n");
 		return -1;
