@@ -21,6 +21,8 @@
  * OF THIS SOFTWARE.
  */
 
+#include <config.h>
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -32,32 +34,54 @@
 
 #include <wayland-client.h>
 #include "../shared/os-compatibility.h"
+#include "xdg-shell-client-protocol.h"
+#include "fullscreen-shell-client-protocol.h"
 
 struct display {
 	struct wl_display *display;
 	struct wl_registry *registry;
 	struct wl_compositor *compositor;
-	struct wl_shell *shell;
+	struct xdg_shell *shell;
+	struct _wl_fullscreen_shell *fshell;
 	struct wl_shm *shm;
 	uint32_t formats;
 };
 
+struct buffer {
+	struct wl_buffer *buffer;
+	void *shm_data;
+	int busy;
+};
+
 struct window {
 	struct display *display;
 	int width, height;
 	struct wl_surface *surface;
-	struct wl_shell_surface *shell_surface;
-	struct wl_buffer *buffer;
-	void *shm_data;
+	struct xdg_surface *xdg_surface;
+	struct buffer buffers[2];
+	struct buffer *prev_buffer;
 	struct wl_callback *callback;
 };
 
-static struct wl_buffer *
-create_shm_buffer(struct display *display,
-		  int width, int height, uint32_t format, void **data_out)
+static int running = 1;
+
+static void
+buffer_release(void *data, struct wl_buffer *buffer)
+{
+	struct buffer *mybuf = data;
+
+	mybuf->busy = 0;
+}
+
+static const struct wl_buffer_listener buffer_listener = {
+	buffer_release
+};
+
+static int
+create_shm_buffer(struct display *display, struct buffer *buffer,
+		  int width, int height, uint32_t format)
 {
 	struct wl_shm_pool *pool;
-	struct wl_buffer *buffer;
 	int fd, size, stride;
 	void *data;
 
@@ -68,83 +92,101 @@ create_shm_buffer(struct display *display,
 	if (fd < 0) {
 		fprintf(stderr, "creating a buffer file for %d B failed: %m\n",
 			size);
-		return NULL;
+		return -1;
 	}
 
 	data = mmap(NULL, size, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
 	if (data == MAP_FAILED) {
 		fprintf(stderr, "mmap failed: %m\n");
 		close(fd);
-		return NULL;
+		return -1;
 	}
 
 	pool = wl_shm_create_pool(display->shm, fd, size);
-	buffer = wl_shm_pool_create_buffer(pool, 0,
-					   width, height, stride, format);
+	buffer->buffer = wl_shm_pool_create_buffer(pool, 0,
+						   width, height,
+						   stride, format);
+	wl_buffer_add_listener(buffer->buffer, &buffer_listener, buffer);
 	wl_shm_pool_destroy(pool);
 	close(fd);
 
-	*data_out = data;
+	buffer->shm_data = data;
 
-	return buffer;
+	return 0;
 }
 
 static void
-handle_ping(void *data, struct wl_shell_surface *shell_surface,
-							uint32_t serial)
-{
-	wl_shell_surface_pong(shell_surface, serial);
-}
-
-static void
-handle_configure(void *data, struct wl_shell_surface *shell_surface,
-		 uint32_t edges, int32_t width, int32_t height)
+handle_configure(void *data, struct xdg_surface *surface,
+		 int32_t width, int32_t height)
 {
 }
 
 static void
-handle_popup_done(void *data, struct wl_shell_surface *shell_surface)
+handle_change_state(void *data, struct xdg_surface *xdg_surface,
+		    uint32_t state,
+		    uint32_t value,
+		    uint32_t serial)
 {
 }
 
-static const struct wl_shell_surface_listener shell_surface_listener = {
-	handle_ping,
+static void
+handle_activated(void *data, struct xdg_surface *xdg_surface)
+{
+}
+
+static void
+handle_deactivated(void *data, struct xdg_surface *xdg_surface)
+{
+}
+
+static void
+handle_delete(void *data, struct xdg_surface *xdg_surface)
+{
+	running = 0;
+}
+
+static const struct xdg_surface_listener xdg_surface_listener = {
 	handle_configure,
-	handle_popup_done
+	handle_change_state,
+	handle_activated,
+	handle_deactivated,
+	handle_delete,
 };
 
 static struct window *
 create_window(struct display *display, int width, int height)
 {
 	struct window *window;
-	
-	window = malloc(sizeof *window);
 
-	window->buffer = create_shm_buffer(display,
-					   width, height,
-					   WL_SHM_FORMAT_XRGB8888,
-					   &window->shm_data);
-
-	if (!window->buffer) {
-		free(window);
+	window = calloc(1, sizeof *window);
+	if (!window)
 		return NULL;
-	}
 
 	window->callback = NULL;
 	window->display = display;
 	window->width = width;
 	window->height = height;
 	window->surface = wl_compositor_create_surface(display->compositor);
-	window->shell_surface = wl_shell_get_shell_surface(display->shell,
-							   window->surface);
 
-	if (window->shell_surface)
-		wl_shell_surface_add_listener(window->shell_surface,
-					      &shell_surface_listener, window);
+	if (display->shell) {
+		window->xdg_surface =
+			xdg_shell_get_xdg_surface(display->shell,
+						  window->surface);
 
-	wl_shell_surface_set_title(window->shell_surface, "simple-shm");
+		assert(window->xdg_surface);
 
-	wl_shell_surface_set_toplevel(window->shell_surface);
+		xdg_surface_add_listener(window->xdg_surface,
+					 &xdg_surface_listener, window);
+
+		xdg_surface_set_title(window->xdg_surface, "simple-shm");
+	} else if (display->fshell) {
+		_wl_fullscreen_shell_present_surface(display->fshell,
+						     window->surface,
+						     _WL_FULLSCREEN_SHELL_PRESENT_METHOD_DEFAULT,
+						     NULL);
+	} else {
+		assert(0);
+	}
 
 	return window;
 }
@@ -155,12 +197,46 @@ destroy_window(struct window *window)
 	if (window->callback)
 		wl_callback_destroy(window->callback);
 
-	wl_buffer_destroy(window->buffer);
-	wl_shell_surface_destroy(window->shell_surface);
+	if (window->buffers[0].buffer)
+		wl_buffer_destroy(window->buffers[0].buffer);
+	if (window->buffers[1].buffer)
+		wl_buffer_destroy(window->buffers[1].buffer);
+
+	if (window->xdg_surface)
+		xdg_surface_destroy(window->xdg_surface);
 	wl_surface_destroy(window->surface);
 	free(window);
 }
 
+static struct buffer *
+window_next_buffer(struct window *window)
+{
+	struct buffer *buffer;
+	int ret = 0;
+
+	if (!window->buffers[0].busy)
+		buffer = &window->buffers[0];
+	else if (!window->buffers[1].busy)
+		buffer = &window->buffers[1];
+	else
+		return NULL;
+
+	if (!buffer->buffer) {
+		ret = create_shm_buffer(window->display, buffer,
+					window->width, window->height,
+					WL_SHM_FORMAT_XRGB8888);
+
+		if (ret < 0)
+			return NULL;
+
+		/* paint the padding */
+		memset(buffer->shm_data, 0xff,
+		       window->width * window->height * 4);
+	}
+
+	return buffer;
+}
+
 static void
 paint_pixels(void *image, int padding, int width, int height, uint32_t time)
 {
@@ -213,8 +289,19 @@ static void
 redraw(void *data, struct wl_callback *callback, uint32_t time)
 {
 	struct window *window = data;
+	struct buffer *buffer;
 
-	paint_pixels(window->shm_data, 20, window->width, window->height, time);
+	buffer = window_next_buffer(window);
+	if (!buffer) {
+		fprintf(stderr,
+			!callback ? "Failed to create the first buffer.\n" :
+			"Both buffers busy at redraw(). Server bug?\n");
+		abort();
+	}
+
+	paint_pixels(buffer->shm_data, 20, window->width, window->height, time);
+
+	wl_surface_attach(window->surface, buffer->buffer, 0, 0);
 	wl_surface_damage(window->surface,
 			  20, 20, window->width - 40, window->height - 40);
 
@@ -224,6 +311,7 @@ redraw(void *data, struct wl_callback *callback, uint32_t time)
 	window->callback = wl_surface_frame(window->surface);
 	wl_callback_add_listener(window->callback, &frame_listener, window);
 	wl_surface_commit(window->surface);
+	buffer->busy = 1;
 }
 
 static const struct wl_callback_listener frame_listener = {
@@ -238,10 +326,26 @@ shm_format(void *data, struct wl_shm *wl_shm, uint32_t format)
 	d->formats |= (1 << format);
 }
 
-struct wl_shm_listener shm_listenter = {
+struct wl_shm_listener shm_listener = {
 	shm_format
 };
 
+static void
+xdg_shell_ping(void *data, struct xdg_shell *shell, uint32_t serial)
+{
+	xdg_shell_pong(shell, serial);
+}
+
+static const struct xdg_shell_listener xdg_shell_listener = {
+	xdg_shell_ping,
+};
+
+#define XDG_VERSION 3 /* The version of xdg-shell that we implement */
+#ifdef static_assert
+static_assert(XDG_VERSION == XDG_SHELL_VERSION_CURRENT,
+	      "Interface version doesn't match implementation version");
+#endif
+
 static void
 registry_handle_global(void *data, struct wl_registry *registry,
 		       uint32_t id, const char *interface, uint32_t version)
@@ -252,18 +356,30 @@ registry_handle_global(void *data, struct wl_registry *registry,
 		d->compositor =
 			wl_registry_bind(registry,
 					 id, &wl_compositor_interface, 1);
-	} else if (strcmp(interface, "wl_shell") == 0) {
+	} else if (strcmp(interface, "xdg_shell") == 0) {
 		d->shell = wl_registry_bind(registry,
-					    id, &wl_shell_interface, 1);
+					    id, &xdg_shell_interface, 1);
+		xdg_shell_use_unstable_version(d->shell, XDG_VERSION);
+		xdg_shell_add_listener(d->shell, &xdg_shell_listener, d);
+	} else if (strcmp(interface, "_wl_fullscreen_shell") == 0) {
+		d->fshell = wl_registry_bind(registry,
+					     id, &_wl_fullscreen_shell_interface, 1);
 	} else if (strcmp(interface, "wl_shm") == 0) {
 		d->shm = wl_registry_bind(registry,
 					  id, &wl_shm_interface, 1);
-		wl_shm_add_listener(d->shm, &shm_listenter, d);
+		wl_shm_add_listener(d->shm, &shm_listener, d);
 	}
 }
 
+static void
+registry_handle_global_remove(void *data, struct wl_registry *registry,
+			      uint32_t name)
+{
+}
+
 static const struct wl_registry_listener registry_listener = {
-	registry_handle_global
+	registry_handle_global,
+	registry_handle_global_remove
 };
 
 static struct display *
@@ -272,6 +388,10 @@ create_display(void)
 	struct display *display;
 
 	display = malloc(sizeof *display);
+	if (display == NULL) {
+		fprintf(stderr, "out of memory\n");
+		exit(1);
+	}
 	display->display = wl_display_connect(NULL);
 	assert(display->display);
 
@@ -304,18 +424,20 @@ destroy_display(struct display *display)
 		wl_shm_destroy(display->shm);
 
 	if (display->shell)
-		wl_shell_destroy(display->shell);
+		xdg_shell_destroy(display->shell);
+
+	if (display->fshell)
+		_wl_fullscreen_shell_release(display->fshell);
 
 	if (display->compositor)
 		wl_compositor_destroy(display->compositor);
 
+	wl_registry_destroy(display->registry);
 	wl_display_flush(display->display);
 	wl_display_disconnect(display->display);
 	free(display);
 }
 
-static int running = 1;
-
 static void
 signal_int(int signum)
 {
@@ -340,8 +462,9 @@ main(int argc, char **argv)
 	sigint.sa_flags = SA_RESETHAND;
 	sigaction(SIGINT, &sigint, NULL);
 
-	memset(window->shm_data, 0xff, window->width * window->height * 4);
-	wl_surface_attach(window->surface, window->buffer, 0, 0);
+	/* Initialise damage to full surface, so the padding gets painted */
+	wl_surface_damage(window->surface, 0, 0,
+			  window->width, window->height);
 
 	redraw(window, NULL, 0);
 
