@@ -23,10 +23,13 @@
 #ifndef _WINDOW_H_
 #define _WINDOW_H_
 
+#include "config.h"
+
 #include <xkbcommon/xkbcommon.h>
 #include <wayland-client.h>
 #include <cairo.h>
 #include "../shared/config-parser.h"
+#include "../shared/zalloc.h"
 
 #define ARRAY_LENGTH(a) (sizeof (a) / sizeof (a)[0])
 
@@ -52,8 +55,19 @@ struct rectangle {
 	int32_t height;
 };
 
+void *
+fail_on_null(void *p);
+void *
+xmalloc(size_t s);
+void *
+xzalloc(size_t s);
+char *
+xstrdup(const char *s);
+void *
+xrealloc(char *p, size_t s);
+
 struct display *
-display_create(int argc, char *argv[]);
+display_create(int *argc, char *argv[]);
 
 void
 display_destroy(struct display *display);
@@ -67,12 +81,15 @@ display_get_user_data(struct display *display);
 struct wl_display *
 display_get_display(struct display *display);
 
+int
+display_has_subcompositor(struct display *display);
+
+cairo_device_t *
+display_get_cairo_device(struct display *display);
+
 struct wl_compositor *
 display_get_compositor(struct display *display);
 
-struct wl_shell *
-display_get_shell(struct display *display);
-
 struct output *
 display_get_output(struct display *display);
 
@@ -87,6 +104,9 @@ typedef void (*display_global_handler_t)(struct display *display,
 void
 display_set_global_handler(struct display *display,
 			   display_global_handler_t handler);
+void
+display_set_global_handler_remove(struct display *display,
+			   display_global_handler_t remove_handler);
 void *
 display_bind(struct display *display, uint32_t name,
 	     const struct wl_interface *interface, uint32_t version);
@@ -126,6 +146,10 @@ display_release_window_surface(struct display *display,
 #define SURFACE_OPAQUE 0x01
 #define SURFACE_SHM    0x02
 
+#define SURFACE_HINT_RESIZE 0x10
+
+#define SURFACE_HINT_RGB565 0x100
+
 cairo_surface_t *
 display_create_surface(struct display *display,
 		       struct wl_surface *surface,
@@ -190,9 +214,12 @@ typedef void (*window_drop_handler_t)(struct window *window,
 				      struct input *input,
 				      int32_t x, int32_t y, void *data);
 
-typedef void (*window_close_handler_t)(struct window *window, void *data);
+typedef void (*window_close_handler_t)(void *data);
 typedef void (*window_fullscreen_handler_t)(struct window *window, void *data);
 
+typedef void (*window_output_handler_t)(struct window *window, struct output *output,
+					int enter, void *data);
+
 typedef void (*widget_resize_handler_t)(struct widget *widget,
 					int32_t width, int32_t height,
 					void *data);
@@ -211,6 +238,31 @@ typedef void (*widget_button_handler_t)(struct widget *widget,
 					uint32_t button,
 					enum wl_pointer_button_state state,
 					void *data);
+typedef void (*widget_touch_down_handler_t)(struct widget *widget,
+					    struct input *input,
+					    uint32_t serial,
+					    uint32_t time,
+					    int32_t id,
+					    float x,
+					    float y,
+					    void *data);
+typedef void (*widget_touch_up_handler_t)(struct widget *widget,
+					  struct input *input,
+					  uint32_t serial,
+					  uint32_t time,
+					  int32_t id,
+					  void *data);
+typedef void (*widget_touch_motion_handler_t)(struct widget *widget,
+					      struct input *input,
+					      uint32_t time,
+					      int32_t id,
+					      float x,
+					      float y,
+					      void *data);
+typedef void (*widget_touch_frame_handler_t)(struct widget *widget, 
+					     struct input *input, void *data);
+typedef void (*widget_touch_cancel_handler_t)(struct widget *widget, 
+					      struct input *input, void *data);
 typedef void (*widget_axis_handler_t)(struct widget *widget,
 				      struct input *input, uint32_t time,
 				      uint32_t axis,
@@ -220,15 +272,18 @@ typedef void (*widget_axis_handler_t)(struct widget *widget,
 struct window *
 window_create(struct display *display);
 struct window *
-window_create_transient(struct display *display, struct window *parent,
-			int32_t x, int32_t y, uint32_t flags);
-struct window *
 window_create_custom(struct display *display);
 
+void
+window_set_transient_for(struct window *window, struct window *parent_window);
+struct window *
+window_get_transient_for(struct window *window);
+
 int
 window_has_focus(struct window *window);
 
-typedef void (*menu_func_t)(struct window *window, int index, void *data);
+typedef void (*menu_func_t)(struct window *window,
+			    struct input *input, int index, void *data);
 
 void
 window_show_menu(struct display *display,
@@ -240,12 +295,38 @@ void
 window_show_frame_menu(struct window *window,
 		       struct input *input, uint32_t time);
 
+int
+window_get_buffer_transform(struct window *window);
+
+void
+window_set_buffer_transform(struct window *window,
+			    enum wl_output_transform transform);
+
+uint32_t
+window_get_buffer_scale(struct window *window);
+
+void
+window_set_buffer_scale(struct window *window,
+                        int32_t scale);
+
+uint32_t
+window_get_output_scale(struct window *window);
+
 void
 window_destroy(struct window *window);
 
 struct widget *
 window_add_widget(struct window *window, void *data);
 
+enum subsurface_mode {
+	SUBSURFACE_SYNCHRONIZED,
+	SUBSURFACE_DESYNCHRONIZED
+};
+
+struct widget *
+window_add_subsurface(struct window *window, void *data,
+		      enum subsurface_mode default_mode);
+
 typedef void (*data_func_t)(void *data, size_t len,
 			    int32_t x, int32_t y, void *user_data);
 
@@ -270,17 +351,8 @@ window_get_surface(struct window *window);
 struct wl_surface *
 window_get_wl_surface(struct window *window);
 
-struct wl_shell_surface *
-window_get_wl_shell_surface(struct window *window);
-
-void
-window_flush(struct window *window);
-
-void
-window_set_surface(struct window *window, cairo_surface_t *surface);
-
-void
-window_create_surface(struct window *window);
+struct wl_subsurface *
+widget_get_wl_subsurface(struct widget *widget);
 
 enum window_buffer_type {
 	WINDOW_BUFFER_TYPE_EGL_WINDOW,
@@ -294,6 +366,9 @@ display_surface_damage(struct display *display, cairo_surface_t *cairo_surface,
 void
 window_set_buffer_type(struct window *window, enum window_buffer_type type);
 
+enum window_buffer_type
+window_get_buffer_type(struct window *window);
+
 int
 window_is_fullscreen(struct window *window);
 
@@ -306,6 +381,9 @@ window_is_maximized(struct window *window);
 void
 window_set_maximized(struct window *window, int maximized);
 
+void
+window_set_minimized(struct window *window);
+
 void
 window_set_user_data(struct window *window, void *data);
 
@@ -334,6 +412,9 @@ window_set_close_handler(struct window *window,
 void
 window_set_fullscreen_handler(struct window *window,
 			      window_fullscreen_handler_t handler);
+void
+window_set_output_handler(struct window *window,
+			  window_output_handler_t handler);
 
 void
 window_set_title(struct window *window, const char *title);
@@ -344,6 +425,15 @@ window_get_title(struct window *window);
 void
 window_set_text_cursor_position(struct window *window, int32_t x, int32_t y);
 
+enum preferred_format {
+	WINDOW_PREFERRED_FORMAT_NONE,
+	WINDOW_PREFERRED_FORMAT_RGB565
+};
+
+void
+window_set_preferred_format(struct window *window,
+			    enum preferred_format format);
+
 int
 widget_set_tooltip(struct widget *parent, char *entry, float x, float y);
 
@@ -356,6 +446,8 @@ widget_add_widget(struct widget *parent, void *data);
 void
 widget_destroy(struct widget *widget);
 void
+widget_set_default_cursor(struct widget *widget, int cursor);
+void
 widget_get_allocation(struct widget *widget, struct rectangle *allocation);
 
 void
@@ -371,6 +463,18 @@ widget_schedule_resize(struct widget *widget, int32_t width, int32_t height);
 void *
 widget_get_user_data(struct widget *widget);
 
+cairo_t *
+widget_cairo_create(struct widget *widget);
+
+struct wl_surface *
+widget_get_wl_surface(struct widget *widget);
+
+uint32_t
+widget_get_last_time(struct widget *widget);
+
+void
+widget_input_region_add(struct widget *widget, const struct rectangle *rect);
+
 void
 widget_set_redraw_handler(struct widget *widget,
 			  widget_redraw_handler_t handler);
@@ -390,16 +494,34 @@ void
 widget_set_button_handler(struct widget *widget,
 			  widget_button_handler_t handler);
 void
+widget_set_touch_down_handler(struct widget *widget,
+			      widget_touch_down_handler_t handler);
+void
+widget_set_touch_up_handler(struct widget *widget,
+			    widget_touch_up_handler_t handler);
+void
+widget_set_touch_motion_handler(struct widget *widget,
+				widget_touch_motion_handler_t handler);
+void
+widget_set_touch_frame_handler(struct widget *widget,
+			       widget_touch_frame_handler_t handler);
+void
+widget_set_touch_cancel_handler(struct widget *widget,
+				widget_touch_cancel_handler_t handler);
+void
 widget_set_axis_handler(struct widget *widget,
 			widget_axis_handler_t handler);
-
 void
 widget_schedule_redraw(struct widget *widget);
+void
+widget_set_use_cairo(struct widget *widget, int use_cairo);
 
 struct widget *
-frame_create(struct window *window, void *data);
+window_frame_create(struct window *window, void *data);
+
 void
-frame_set_child_size(struct widget *widget, int child_width, int child_height);
+window_frame_set_child_size(struct widget *widget, int child_width,
+			    int child_height);
 
 void
 input_set_pointer_image(struct input *input, int pointer);
@@ -407,6 +529,9 @@ input_set_pointer_image(struct input *input, int pointer);
 void
 input_get_position(struct input *input, int32_t *x, int32_t *y);
 
+int
+input_get_touch(struct input *input, int32_t id, float *x, float *y);
+
 #define MOD_SHIFT_MASK		0x01
 #define MOD_ALT_MASK		0x02
 #define MOD_CONTROL_MASK	0x04
@@ -414,6 +539,12 @@ input_get_position(struct input *input, int32_t *x, int32_t *y);
 uint32_t
 input_get_modifiers(struct input *input);
 
+void
+touch_grab(struct input *input, int32_t touch_id);
+
+void
+touch_ungrab(struct input *input);
+
 void
 input_grab(struct input *input, struct widget *widget, uint32_t button);
 
@@ -443,6 +574,9 @@ input_accept(struct input *input, const char *type);
 void
 input_receive_drag_data(struct input *input, const char *mime_type,
 			data_func_t func, void *user_data);
+int
+input_receive_drag_data_to_fd(struct input *input,
+			      const char *mime_type, int fd);
 
 int
 input_receive_selection_data(struct input *input, const char *mime_type,
@@ -467,4 +601,24 @@ output_get_allocation(struct output *output, struct rectangle *allocation);
 struct wl_output *
 output_get_wl_output(struct output *output);
 
+enum wl_output_transform
+output_get_transform(struct output *output);
+
+uint32_t
+output_get_scale(struct output *output);
+
+const char *
+output_get_make(struct output *output);
+
+const char *
+output_get_model(struct output *output);
+
+void
+keysym_modifiers_add(struct wl_array *modifiers_map,
+		     const char *name);
+
+xkb_mod_mask_t
+keysym_modifiers_get_mask(struct wl_array *modifiers_map,
+			  const char *name);
+
 #endif
