@@ -20,6 +20,8 @@
  * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  */
 
+#include "config.h"
+
 #include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
@@ -34,7 +36,6 @@
 #include "../wcap/wcap-decode.h"
 
 struct screenshooter {
-	struct wl_object base;
 	struct weston_compositor *ec;
 	struct wl_global *global;
 	struct wl_client *client;
@@ -44,8 +45,9 @@ struct screenshooter {
 
 struct screenshooter_frame_listener {
 	struct wl_listener listener;
-	struct wl_buffer *buffer;
-	struct wl_resource *resource;
+	struct weston_buffer *buffer;
+	weston_screenshooter_done_func_t done;
+	void *data;
 };
 
 static void
@@ -61,6 +63,13 @@ copy_bgra_yflip(uint8_t *dst, uint8_t *src, int height, int stride)
 	}
 }
 
+static void
+copy_bgra(uint8_t *dst, uint8_t *src, int height, int stride)
+{
+	/* TODO: optimize this out */
+	memcpy(dst, src, height * stride);
+}
+
 static void
 copy_row_swap_RB(void *vdst, void *vsrc, int bytes)
 {
@@ -91,6 +100,19 @@ copy_rgba_yflip(uint8_t *dst, uint8_t *src, int height, int stride)
 	}
 }
 
+static void
+copy_rgba(uint8_t *dst, uint8_t *src, int height, int stride)
+{
+	uint8_t *end;
+
+	end = dst + height * stride;
+	while (dst < end) {
+		copy_row_swap_RB(dst, src, stride);
+		dst += stride;
+		src += stride;
+	}
+}
+
 static void
 screenshooter_frame_notify(struct wl_listener *listener, void *data)
 {
@@ -98,76 +120,132 @@ screenshooter_frame_notify(struct wl_listener *listener, void *data)
 		container_of(listener,
 			     struct screenshooter_frame_listener, listener);
 	struct weston_output *output = data;
+	struct weston_compositor *compositor = output->compositor;
 	int32_t stride;
 	uint8_t *pixels, *d, *s;
 
 	output->disable_planes--;
 	wl_list_remove(&listener->link);
-	stride = l->buffer->width * 4;
+	stride = l->buffer->width * (PIXMAN_FORMAT_BPP(compositor->read_format) / 8);
 	pixels = malloc(stride * l->buffer->height);
 
 	if (pixels == NULL) {
-		wl_resource_post_no_memory(l->resource);
+		l->done(l->data, WESTON_SCREENSHOOTER_NO_MEMORY);
 		free(l);
 		return;
 	}
 
-	glPixelStorei(GL_PACK_ALIGNMENT, 1);
-	glReadPixels(0, 0, output->current->width, output->current->height,
-		     output->compositor->read_format,
-		     GL_UNSIGNED_BYTE, pixels);
+	compositor->renderer->read_pixels(output,
+			     compositor->read_format, pixels,
+			     0, 0, output->current_mode->width,
+			     output->current_mode->height);
 
-	stride = wl_shm_buffer_get_stride(l->buffer);
+	stride = wl_shm_buffer_get_stride(l->buffer->shm_buffer);
 
-	d = wl_shm_buffer_get_data(l->buffer);
+	d = wl_shm_buffer_get_data(l->buffer->shm_buffer);
 	s = pixels + stride * (l->buffer->height - 1);
 
-	switch (output->compositor->read_format) {
-	case GL_BGRA_EXT:
-		copy_bgra_yflip(d, s, output->current->height, stride);
+	wl_shm_buffer_begin_access(l->buffer->shm_buffer);
+
+	switch (compositor->read_format) {
+	case PIXMAN_a8r8g8b8:
+	case PIXMAN_x8r8g8b8:
+		if (compositor->capabilities & WESTON_CAP_CAPTURE_YFLIP)
+			copy_bgra_yflip(d, s, output->current_mode->height, stride);
+		else
+			copy_bgra(d, pixels, output->current_mode->height, stride);
 		break;
-	case GL_RGBA:
-		copy_rgba_yflip(d, s, output->current->height, stride);
+	case PIXMAN_x8b8g8r8:
+	case PIXMAN_a8b8g8r8:
+		if (compositor->capabilities & WESTON_CAP_CAPTURE_YFLIP)
+			copy_rgba_yflip(d, s, output->current_mode->height, stride);
+		else
+			copy_rgba(d, pixels, output->current_mode->height, stride);
 		break;
 	default:
 		break;
 	}
 
-	screenshooter_send_done(l->resource);
+	wl_shm_buffer_end_access(l->buffer->shm_buffer);
+
+	l->done(l->data, WESTON_SCREENSHOOTER_SUCCESS);
 	free(pixels);
 	free(l);
 }
 
+WL_EXPORT int
+weston_screenshooter_shoot(struct weston_output *output,
+			   struct weston_buffer *buffer,
+			   weston_screenshooter_done_func_t done, void *data)
+{
+	struct screenshooter_frame_listener *l;
+
+	if (!wl_shm_buffer_get(buffer->resource)) {
+		done(data, WESTON_SCREENSHOOTER_BAD_BUFFER);
+		return -1;
+	}
+
+	buffer->shm_buffer = wl_shm_buffer_get(buffer->resource);
+	buffer->width = wl_shm_buffer_get_width(buffer->shm_buffer);
+	buffer->height = wl_shm_buffer_get_height(buffer->shm_buffer);
+
+	if (buffer->width < output->current_mode->width ||
+	    buffer->height < output->current_mode->height) {
+		done(data, WESTON_SCREENSHOOTER_BAD_BUFFER);
+		return -1;
+	}
+
+	l = malloc(sizeof *l);
+	if (l == NULL) {
+		done(data, WESTON_SCREENSHOOTER_NO_MEMORY);
+		return -1;
+	}
+
+	l->buffer = buffer;
+	l->done = done;
+	l->data = data;
+	l->listener.notify = screenshooter_frame_notify;
+	wl_signal_add(&output->frame_signal, &l->listener);
+	output->disable_planes++;
+	weston_output_schedule_repaint(output);
+
+	return 0;
+}
+
+static void
+screenshooter_done(void *data, enum weston_screenshooter_outcome outcome)
+{
+	struct wl_resource *resource = data;
+
+	switch (outcome) {
+	case WESTON_SCREENSHOOTER_SUCCESS:
+		screenshooter_send_done(resource);
+		break;
+	case WESTON_SCREENSHOOTER_NO_MEMORY:
+		wl_resource_post_no_memory(resource);
+		break;
+	default:
+		break;
+	}
+}
+
 static void
 screenshooter_shoot(struct wl_client *client,
 		    struct wl_resource *resource,
 		    struct wl_resource *output_resource,
 		    struct wl_resource *buffer_resource)
 {
-	struct weston_output *output = output_resource->data;
-	struct screenshooter_frame_listener *l;
-	struct wl_buffer *buffer = buffer_resource->data;
+	struct weston_output *output =
+		wl_resource_get_user_data(output_resource);
+	struct weston_buffer *buffer =
+		weston_buffer_from_resource(buffer_resource);
 
-	if (!wl_buffer_is_shm(buffer))
-		return;
-
-	if (buffer->width < output->current->width ||
-	    buffer->height < output->current->height)
-		return;
-
-	l = malloc(sizeof *l);
-	if (l == NULL) {
+	if (buffer == NULL) {
 		wl_resource_post_no_memory(resource);
 		return;
 	}
 
-	l->buffer = buffer;
-	l->resource = resource;
-
-	l->listener.notify = screenshooter_frame_notify;
-	wl_signal_add(&output->frame_signal, &l->listener);
-	output->disable_planes++;
-	weston_output_schedule_repaint(output);
+	weston_screenshooter_shoot(output, buffer, screenshooter_done, resource);
 }
 
 struct screenshooter_interface screenshooter_implementation = {
@@ -181,14 +259,17 @@ bind_shooter(struct wl_client *client,
 	struct screenshooter *shooter = data;
 	struct wl_resource *resource;
 
-	resource = wl_client_add_object(client, &screenshooter_interface,
-			     &screenshooter_implementation, id, data);
+	resource = wl_resource_create(client,
+				      &screenshooter_interface, 1, id);
 
 	if (client != shooter->client) {
 		wl_resource_post_error(resource, WL_DISPLAY_ERROR_INVALID_OBJECT,
 				       "screenshooter failed: permission denied");
 		wl_resource_destroy(resource);
 	}
+
+	wl_resource_set_implementation(resource, &screenshooter_implementation,
+				       data, NULL);
 }
 
 static void
@@ -201,7 +282,7 @@ screenshooter_sigchld(struct weston_process *process, int status)
 }
 
 static void
-screenshooter_binding(struct wl_seat *seat, uint32_t time, uint32_t key,
+screenshooter_binding(struct weston_seat *seat, uint32_t time, uint32_t key,
 		      void *data)
 {
 	struct screenshooter *shooter = data;
@@ -216,10 +297,11 @@ screenshooter_binding(struct wl_seat *seat, uint32_t time, uint32_t key,
 struct weston_recorder {
 	struct weston_output *output;
 	uint32_t *frame, *rect;
+	uint32_t *tmpbuf;
 	uint32_t total;
 	int fd;
 	struct wl_listener frame_listener;
-	int count;
+	int count, destroying;
 };
 
 static uint32_t *
@@ -254,53 +336,7 @@ component_delta(uint32_t next, uint32_t prev)
 }
 
 static void
-transform_rect(struct weston_output *output, pixman_box32_t *r)
-{
-	pixman_box32_t s = *r;
-
-	switch(output->transform) {
-	case WL_OUTPUT_TRANSFORM_FLIPPED:
-	case WL_OUTPUT_TRANSFORM_FLIPPED_90:
-	case WL_OUTPUT_TRANSFORM_FLIPPED_180:
-	case WL_OUTPUT_TRANSFORM_FLIPPED_270:
-		s.x1 = output->width - r->x2;
-		s.x2 = output->width - r->x1;
-		break;
-	default:
-		break;
-	}
-
-        switch(output->transform) {
-        case WL_OUTPUT_TRANSFORM_NORMAL:
-        case WL_OUTPUT_TRANSFORM_FLIPPED:
-		r->x1 = s.x1;
-		r->x2 = s.x2;
-                break;
-        case WL_OUTPUT_TRANSFORM_90:
-        case WL_OUTPUT_TRANSFORM_FLIPPED_90:
-		r->x1 = output->current->width - s.y2;
-		r->y1 = s.x1;
-		r->x2 = output->current->width - s.y1;
-		r->y2 = s.x2;
-                break;
-        case WL_OUTPUT_TRANSFORM_180:
-        case WL_OUTPUT_TRANSFORM_FLIPPED_180:
-		r->x1 = output->current->width - s.x2;
-		r->y1 = output->current->height - s.y2;
-		r->x2 = output->current->width - s.x1;
-		r->y2 = output->current->height - s.y1;
-                break;
-        case WL_OUTPUT_TRANSFORM_270:
-        case WL_OUTPUT_TRANSFORM_FLIPPED_270:
-		r->x1 = s.y1; 
-		r->y1 = output->current->height - s.x2;
-		r->x2 = s.y2; 
-		r->y2 = output->current->height - s.x1;
-                break;
-        default:
-                break;
-        }
-}
+weston_recorder_destroy(struct weston_recorder *recorder);
 
 static void
 weston_recorder_frame_notify(struct wl_listener *listener, void *data)
@@ -308,9 +344,10 @@ weston_recorder_frame_notify(struct wl_listener *listener, void *data)
 	struct weston_recorder *recorder =
 		container_of(listener, struct weston_recorder, frame_listener);
 	struct weston_output *output = data;
+	struct weston_compositor *compositor = output->compositor;
 	uint32_t msecs = output->frame_time;
 	pixman_box32_t *r;
-	pixman_region32_t damage, *previous_damage;
+	pixman_region32_t damage, transformed_damage;
 	int i, j, k, n, width, height, run, stride;
 	uint32_t delta, prev, *d, *s, *p, next;
 	struct {
@@ -318,26 +355,31 @@ weston_recorder_frame_notify(struct wl_listener *listener, void *data)
 		uint32_t nrects;
 	} header;
 	struct iovec v[2];
+	int do_yflip;
+	int y_orig;
+	uint32_t *outbuf;
 
-	/* When recording, this will be exactly the region that was repainted
-	 * in this frame. Since overlays are disabled, the whole primary plane
-	 * damage is rendered. For the first frame, the whole output will be
-	 * damaged and that damage will be added to both buffers causing the
-	 * non-current buffer damage to be while output. Rendering will clear
-	 * all the damage in the current buffer so in the next frame (when
-	 * that is non-current) the only damage left will be the one added
-	 * from the primary plane. */
-	previous_damage = &output->buffer_damage[output->current_buffer ^ 1];
+	do_yflip = !!(compositor->capabilities & WESTON_CAP_CAPTURE_YFLIP);
+	if (do_yflip)
+		outbuf = recorder->rect;
+	else
+		outbuf = recorder->tmpbuf;
 
 	pixman_region32_init(&damage);
-	pixman_region32_intersect(&damage, &output->region, previous_damage);
+	pixman_region32_init(&transformed_damage);
+	pixman_region32_intersect(&damage, &output->region,
+				  &output->previous_damage);
+	pixman_region32_translate(&damage, -output->x, -output->y);
+	weston_transformed_region(output->width, output->height,
+				 output->transform, output->current_scale,
+				 &damage, &transformed_damage);
+	pixman_region32_fini(&damage);
 
-	r = pixman_region32_rectangles(&damage, &n);
-	if (n == 0)
+	r = pixman_region32_rectangles(&transformed_damage, &n);
+	if (n == 0) {
+		pixman_region32_fini(&transformed_damage);
 		return;
-
-	for (i = 0; i < n; i++)
-		transform_rect(output, &r[i]);
+	}
 
 	header.msecs = msecs;
 	header.nrects = n;
@@ -346,22 +388,31 @@ weston_recorder_frame_notify(struct wl_listener *listener, void *data)
 	v[1].iov_base = r;
 	v[1].iov_len = n * sizeof *r;
 	recorder->total += writev(recorder->fd, v, 2);
-	stride = output->current->width;
+	stride = output->current_mode->width;
 
 	for (i = 0; i < n; i++) {
 		width = r[i].x2 - r[i].x1;
 		height = r[i].y2 - r[i].y1;
-		glReadPixels(r[i].x1, output->current->height - r[i].y2,
-			     width, height,
-			     output->compositor->read_format,
-			     GL_UNSIGNED_BYTE, recorder->rect);
+
+		if (do_yflip)
+			y_orig = output->current_mode->height - r[i].y2;
+		else
+			y_orig = r[i].y1;
+
+		compositor->renderer->read_pixels(output,
+				compositor->read_format, recorder->rect,
+				r[i].x1, y_orig, width, height);
 
 		s = recorder->rect;
-		p = recorder->rect;
+		p = outbuf;
 		run = prev = 0; /* quiet gcc */
 		for (j = 0; j < height; j++) {
-			d = recorder->frame +
-				stride * (r[i].y2 - j - 1) + r[i].x1;
+			if (do_yflip)
+				y_orig = r[i].y2 - j - 1;
+			else
+				y_orig = r[i].y1 + j;
+			d = recorder->frame + stride * y_orig + r[i].x1;
+
 			for (k = 0; k < width; k++) {
 				next = *s++;
 				delta = component_delta(next, *d);
@@ -379,49 +430,87 @@ weston_recorder_frame_notify(struct wl_listener *listener, void *data)
 		p = output_run(p, prev, run);
 
 		recorder->total += write(recorder->fd,
-					 recorder->rect,
-					 (p - recorder->rect) * 4);
+					 outbuf, (p - outbuf) * 4);
 
 #if 0
 		fprintf(stderr,
 			"%dx%d at %d,%d rle from %d to %d bytes (%f) total %dM\n",
 			width, height, r[i].x1, r[i].y1,
-			width * height * 4, (int) (p - recorder->rect) * 4,
-			(float) (p - recorder->rect) / (width * height),
+			width * height * 4, (int) (p - outbuf) * 4,
+			(float) (p - outbuf) / (width * height),
 			recorder->total / 1024 / 1024);
 #endif
 	}
 
-	pixman_region32_fini(&damage);
+	pixman_region32_fini(&transformed_damage);
+	recorder->count++;
+
+	if (recorder->destroying)
+		weston_recorder_destroy(recorder);
+}
+
+static void
+weston_recorder_free(struct weston_recorder *recorder)
+{
+	if (recorder == NULL)
+		return;
+	free(recorder->rect);
+	free(recorder->tmpbuf);
+	free(recorder->frame);
+	free(recorder);
 }
 
 static void
 weston_recorder_create(struct weston_output *output, const char *filename)
 {
+	struct weston_compositor *compositor = output->compositor;
 	struct weston_recorder *recorder;
 	int stride, size;
 	struct { uint32_t magic, format, width, height; } header;
+	int do_yflip;
+
+	do_yflip = !!(compositor->capabilities & WESTON_CAP_CAPTURE_YFLIP);
 
 	recorder = malloc(sizeof *recorder);
+	if (recorder == NULL) {
+		weston_log("%s: out of memory\n", __func__);
+		return;
+	}
 
-	stride = output->current->width;
-	size = stride * 4 * output->current->height;
-	recorder->frame = malloc(size);
+	stride = output->current_mode->width;
+	size = stride * 4 * output->current_mode->height;
+	recorder->frame = zalloc(size);
 	recorder->rect = malloc(size);
 	recorder->total = 0;
 	recorder->count = 0;
+	recorder->destroying = 0;
 	recorder->output = output;
-	memset(recorder->frame, 0, size);
+
+	if ((recorder->frame == NULL) || (recorder->rect == NULL)) {
+		weston_log("%s: out of memory\n", __func__);
+		weston_recorder_free(recorder);
+		return;
+	}
+
+	if (do_yflip)
+		recorder->tmpbuf = NULL;
+	else
+		recorder->tmpbuf = malloc(size);
 
 	header.magic = WCAP_HEADER_MAGIC;
 
-	switch (output->compositor->read_format) {
-	case GL_BGRA_EXT:
+	switch (compositor->read_format) {
+	case PIXMAN_x8r8g8b8:
+	case PIXMAN_a8r8g8b8:
 		header.format = WCAP_FORMAT_XRGB8888;
 		break;
-	case GL_RGBA:
+	case PIXMAN_a8b8g8r8:
 		header.format = WCAP_FORMAT_XBGR8888;
 		break;
+	default:
+		weston_log("unknown recorder format\n");
+		weston_recorder_free(recorder);
+		return;
 	}
 
 	recorder->fd = open(filename,
@@ -429,11 +518,12 @@ weston_recorder_create(struct weston_output *output, const char *filename)
 
 	if (recorder->fd < 0) {
 		weston_log("problem opening output file %s: %m\n", filename);
+		weston_recorder_free(recorder);
 		return;
 	}
 
-	header.width = output->current->width;
-	header.height = output->current->height;
+	header.width = output->current_mode->width;
+	header.height = output->current_mode->height;
 	recorder->total += write(recorder->fd, &header, sizeof header);
 
 	recorder->frame_listener.notify = weston_recorder_frame_notify;
@@ -447,37 +537,47 @@ weston_recorder_destroy(struct weston_recorder *recorder)
 {
 	wl_list_remove(&recorder->frame_listener.link);
 	close(recorder->fd);
-	free(recorder->frame);
-	free(recorder->rect);
 	recorder->output->disable_planes--;
-	free(recorder);
+	weston_recorder_free(recorder);
 }
 
 static void
-recorder_binding(struct wl_seat *seat, uint32_t time, uint32_t key, void *data)
+recorder_binding(struct weston_seat *seat, uint32_t time, uint32_t key, void *data)
 {
 	struct weston_seat *ws = (struct weston_seat *) seat;
 	struct weston_compositor *ec = ws->compositor;
-	struct weston_output *output =
-		container_of(ec->output_list.next,
-			     struct weston_output, link);
-	struct wl_listener *listener;
+	struct weston_output *output;
+	struct wl_listener *listener = NULL;
 	struct weston_recorder *recorder;
 	static const char filename[] = "capture.wcap";
 
-	listener = wl_signal_get(&output->frame_signal,
-				 weston_recorder_frame_notify);
+	wl_list_for_each(output, &seat->compositor->output_list, link) {
+		listener = wl_signal_get(&output->frame_signal,
+					 weston_recorder_frame_notify);
+		if (listener)
+			break;
+	}
+
 	if (listener) {
 		recorder = container_of(listener, struct weston_recorder,
 					frame_listener);
 
-		fprintf(stderr,
+		weston_log(
 			"stopping recorder, total file size %dM, %d frames\n",
 			recorder->total / (1024 * 1024), recorder->count);
 
-		weston_recorder_destroy(recorder);
+		recorder->destroying = 1;
+		weston_output_schedule_repaint(recorder->output);
 	} else {
-		fprintf(stderr, "starting recorder, file %s\n", filename);
+		if (seat->keyboard && seat->keyboard->focus &&
+		    seat->keyboard->focus->output)
+			output = seat->keyboard->focus->output;
+		else
+			output = container_of(ec->output_list.next,
+					      struct weston_output, link);
+
+		weston_log("starting recorder for output %s, file %s\n",
+			   output->name, filename);
 		weston_recorder_create(output, filename);
 	}
 }
@@ -488,11 +588,11 @@ screenshooter_destroy(struct wl_listener *listener, void *data)
 	struct screenshooter *shooter =
 		container_of(listener, struct screenshooter, destroy_listener);
 
-	wl_display_remove_global(shooter->ec->wl_display, shooter->global);
+	wl_global_destroy(shooter->global);
 	free(shooter);
 }
 
-void
+WL_EXPORT void
 screenshooter_create(struct weston_compositor *ec)
 {
 	struct screenshooter *shooter;
@@ -501,15 +601,12 @@ screenshooter_create(struct weston_compositor *ec)
 	if (shooter == NULL)
 		return;
 
-	shooter->base.interface = &screenshooter_interface;
-	shooter->base.implementation =
-		(void(**)(void)) &screenshooter_implementation;
 	shooter->ec = ec;
 	shooter->client = NULL;
 
-	shooter->global = wl_display_add_global(ec->wl_display,
-						&screenshooter_interface,
-						shooter, bind_shooter);
+	shooter->global = wl_global_create(ec->wl_display,
+					   &screenshooter_interface, 1,
+					   shooter, bind_shooter);
 	weston_compositor_add_key_binding(ec, KEY_S, MODIFIER_SUPER,
 					  screenshooter_binding, shooter);
 	weston_compositor_add_key_binding(ec, KEY_R, MODIFIER_SUPER,
