@@ -23,6 +23,8 @@
 #ifndef _WESTON_TEST_CLIENT_HELPER_H_
 #define _WESTON_TEST_CLIENT_HELPER_H_
 
+#include "config.h"
+
 #include <assert.h>
 #include "weston-test-runner.h"
 #include "wayland-test-client-protocol.h"
@@ -51,6 +53,7 @@ struct test {
 	struct wl_test *wl_test;
 	int pointer_x;
 	int pointer_y;
+	uint32_t n_egl_buffers;
 };
 
 struct input {
@@ -101,6 +104,9 @@ struct surface {
 struct client *
 client_create(int x, int y, int width, int height);
 
+struct wl_buffer *
+create_shm_buffer(struct client *client, int width, int height, void **pixels);
+
 int
 surface_contains(struct surface *surface, int x, int y);
 
@@ -111,4 +117,16 @@ move_client(struct client *client, int x, int y);
 	assert(wl_display_roundtrip((c)->wl_display) >= 0); \
 } while (0)
 
+struct wl_callback *
+frame_callback_set(struct wl_surface *surface, int *done);
+
+void
+frame_callback_wait(struct client *client, int *done);
+
+int
+get_n_egl_buffers(struct client *client);
+
+void
+skip(const char *fmt, ...);
+
 #endif
