@@ -20,6 +20,8 @@
  * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  */
 
+#include "config.h"
+
 #include <stdlib.h>
 #include <string.h>
 #include <assert.h>
@@ -28,6 +30,11 @@
 #include "../src/compositor.h"
 #include "wayland-test-server-protocol.h"
 
+#ifdef ENABLE_EGL
+#include <EGL/egl.h>
+#include <EGL/eglext.h>
+#endif /* ENABLE_EGL */
+
 struct weston_test {
 	struct weston_compositor *compositor;
 	struct weston_layer layer;
@@ -36,6 +43,7 @@ struct weston_test {
 
 struct weston_test_surface {
 	struct weston_surface *surface;
+	struct weston_view *view;
 	int32_t x, y;
 	struct weston_test *test;
 };
@@ -46,6 +54,12 @@ test_client_sigchld(struct weston_process *process, int status)
 	struct weston_test *test =
 		container_of(process, struct weston_test, process);
 
+	/* Chain up from weston-test-runner's exit code so that automake
+	 * knows the exit status and can report e.g. skipped tests. */
+	if (WIFEXITED(status) && WEXITSTATUS(status) != 0)
+		exit(WEXITSTATUS(status));
+
+	/* In case the child aborted or segfaulted... */
 	assert(status == 0);
 
 	wl_display_terminate(test->compositor->wl_display);
@@ -68,7 +82,7 @@ static void
 notify_pointer_position(struct weston_test *test, struct wl_resource *resource)
 {
 	struct weston_seat *seat = get_seat(test);
-	struct wl_pointer *pointer = seat->seat.pointer;
+	struct weston_pointer *pointer = seat->pointer;
 
 	wl_test_send_pointer_position(resource, pointer->x, pointer->y);
 }
@@ -76,21 +90,17 @@ notify_pointer_position(struct weston_test *test, struct wl_resource *resource)
 static void
 test_surface_configure(struct weston_surface *surface, int32_t sx, int32_t sy)
 {
-	struct weston_test_surface *test_surface = surface->private;
+	struct weston_test_surface *test_surface = surface->configure_private;
 	struct weston_test *test = test_surface->test;
 
-	if (wl_list_empty(&surface->layer_link))
-		wl_list_insert(&test->layer.surface_list,
-			       &surface->layer_link);
+	if (wl_list_empty(&test_surface->view->layer_link))
+		wl_list_insert(&test->layer.view_list,
+			       &test_surface->view->layer_link);
 
-	surface->geometry.x = test_surface->x;
-	surface->geometry.y = test_surface->y;
-	surface->geometry.width = surface->buffer->width;
-	surface->geometry.height = surface->buffer->height;
-	surface->geometry.dirty = 1;
+	weston_view_set_position(test_surface->view,
+				 test_surface->x, test_surface->y);
 
-	if (!weston_surface_is_mapped(surface))
-		weston_surface_update_transform(surface);
+	weston_view_update_transform(test_surface->view);
 }
 
 static void
@@ -98,20 +108,31 @@ move_surface(struct wl_client *client, struct wl_resource *resource,
 	     struct wl_resource *surface_resource,
 	     int32_t x, int32_t y)
 {
-	struct weston_surface *surface = surface_resource->data;
+	struct weston_surface *surface =
+		wl_resource_get_user_data(surface_resource);
 	struct weston_test_surface *test_surface;
 
-	surface->configure = test_surface_configure;
-	if (surface->private == NULL)
-		surface->private = malloc(sizeof *test_surface);
-	test_surface = surface->private;
-	if (test_surface == NULL) {
-		wl_resource_post_no_memory(resource);
-		return;
+	test_surface = surface->configure_private;
+	if (!test_surface) {
+		test_surface = malloc(sizeof *test_surface);
+		if (!test_surface) {
+			wl_resource_post_no_memory(resource);
+			return;
+		}
+
+		test_surface->view = weston_view_create(surface);
+		if (!test_surface->view) {
+			wl_resource_post_no_memory(resource);
+			free(test_surface);
+			return;
+		}
+
+		surface->configure_private = test_surface;
+		surface->configure = test_surface_configure;
 	}
 
 	test_surface->surface = surface;
-	test_surface->test = resource->data;
+	test_surface->test = wl_resource_get_user_data(resource);
 	test_surface->x = x;
 	test_surface->y = y;
 }
@@ -120,12 +141,13 @@ static void
 move_pointer(struct wl_client *client, struct wl_resource *resource,
 	     int32_t x, int32_t y)
 {
-	struct weston_test *test = resource->data;
+	struct weston_test *test = wl_resource_get_user_data(resource);
 	struct weston_seat *seat = get_seat(test);
+	struct weston_pointer *pointer = seat->pointer;
 
-	test->compositor->focus = 1;
-
-	notify_motion(seat, 100, wl_fixed_from_int(x), wl_fixed_from_int(y));
+	notify_motion(seat, 100,
+		      wl_fixed_from_int(x) - pointer->x,
+		      wl_fixed_from_int(y) - pointer->y);
 
 	notify_pointer_position(test, resource);
 }
@@ -134,11 +156,9 @@ static void
 send_button(struct wl_client *client, struct wl_resource *resource,
 	    int32_t button, uint32_t state)
 {
-	struct weston_test *test = resource->data;
+	struct weston_test *test = wl_resource_get_user_data(resource);
 	struct weston_seat *seat = get_seat(test);
 
-	test->compositor->focus = 1;
-
 	notify_button(seat, 100, button, state);
 }
 
@@ -147,15 +167,15 @@ activate_surface(struct wl_client *client, struct wl_resource *resource,
 		 struct wl_resource *surface_resource)
 {
 	struct weston_surface *surface = surface_resource ?
-					surface_resource->data : NULL;
-	struct weston_test *test = resource->data;
+		wl_resource_get_user_data(surface_resource) : NULL;
+	struct weston_test *test = wl_resource_get_user_data(resource);
 	struct weston_seat *seat;
 
 	seat = get_seat(test);
 
 	if (surface) {
 		weston_surface_activate(surface, seat);
-		notify_keyboard_focus_in(seat, &seat->keyboard.keys,
+		notify_keyboard_focus_in(seat, &seat->keyboard->keys,
 					 STATE_UPDATE_AUTOMATIC);
 	}
 	else {
@@ -168,20 +188,60 @@ static void
 send_key(struct wl_client *client, struct wl_resource *resource,
 	 uint32_t key, enum wl_keyboard_key_state state)
 {
-	struct weston_test *test = resource->data;
+	struct weston_test *test = wl_resource_get_user_data(resource);
 	struct weston_seat *seat = get_seat(test);
 
-	test->compositor->focus = 1;
-
 	notify_key(seat, 100, key, state, STATE_UPDATE_AUTOMATIC);
 }
 
+#ifdef ENABLE_EGL
+static int
+is_egl_buffer(struct wl_resource *resource)
+{
+	PFNEGLQUERYWAYLANDBUFFERWL query_buffer =
+		(void *) eglGetProcAddress("eglQueryWaylandBufferWL");
+	EGLint format;
+
+	if (query_buffer(eglGetCurrentDisplay(),
+			 resource,
+			 EGL_TEXTURE_FORMAT,
+			 &format))
+		return 1;
+
+	return 0;
+}
+#endif /* ENABLE_EGL */
+
+static void
+get_n_buffers(struct wl_client *client, struct wl_resource *resource)
+{
+	int n_buffers = 0;
+
+#ifdef ENABLE_EGL
+	struct wl_resource *buffer_resource;
+	int i;
+
+	for (i = 0; i < 1000; i++) {
+		buffer_resource = wl_client_get_object(client, i);
+
+		if (buffer_resource == NULL)
+			continue;
+
+		if (is_egl_buffer(buffer_resource))
+			n_buffers++;
+	}
+#endif /* ENABLE_EGL */
+
+	wl_test_send_n_egl_buffers(resource, n_buffers);
+}
+
 static const struct wl_test_interface test_implementation = {
 	move_surface,
 	move_pointer,
 	send_button,
 	activate_surface,
-	send_key
+	send_key,
+	get_n_buffers,
 };
 
 static void
@@ -190,8 +250,9 @@ bind_test(struct wl_client *client, void *data, uint32_t version, uint32_t id)
 	struct weston_test *test = data;
 	struct wl_resource *resource;
 
-	resource = wl_client_add_object(client, &wl_test_interface,
-					&test_implementation, id, test);
+	resource = wl_resource_create(client, &wl_test_interface, 1, id);
+	wl_resource_set_implementation(resource,
+				       &test_implementation, test, NULL);
 
 	notify_pointer_position(test, resource);
 }
@@ -206,7 +267,7 @@ idle_launch_client(void *data)
 
 	path = getenv("WESTON_TEST_CLIENT_PATH");
 	if (path == NULL)
-		exit(EXIT_FAILURE);
+		return;
 	pid = fork();
 	if (pid == -1)
 		exit(EXIT_FAILURE);
@@ -224,21 +285,21 @@ idle_launch_client(void *data)
 }
 
 WL_EXPORT int
-module_init(struct weston_compositor *ec)
+module_init(struct weston_compositor *ec,
+	    int *argc, char *argv[])
 {
 	struct weston_test *test;
 	struct wl_event_loop *loop;
 
-	test = malloc(sizeof *test);
+	test = zalloc(sizeof *test);
 	if (test == NULL)
 		return -1;
 
-	memset(test, 0, sizeof *test);
 	test->compositor = ec;
 	weston_layer_init(&test->layer, &ec->cursor_layer.link);
 
-	if (wl_display_add_global(ec->wl_display, &wl_test_interface,
-				  test, bind_test) == NULL)
+	if (wl_global_create(ec->wl_display, &wl_test_interface, 1,
+			     test, bind_test) == NULL)
 		return -1;
 
 	loop = wl_display_get_event_loop(ec->wl_display);
